/* FindPanelFix.m created by clindber on Tue 17-Aug-1999 */

#import "RCEnhancePB.h"
#import <ProjectBuilder/PBBundleHost.h>
#import <ProjectBuilder/PBProject.h>
//#import <ProjectBuilder/PBProjectType.h>  // we can't import it!
#import <objc/objc-class.h>

/* 
 * Notifications that PBButtons uses. We'll send them to just so we don't mess up
 * their bundle.
 */
#define MAKE_BUTTON_NOTIFICATION_NAME   @"BuildButtonChangedEnabled"
#define LAUNCH_BUTTON_NOTIFICATION_NAME @"ProcessesButtonChangedEnabled"


/*
 * PBProjectType.h #imports a nonexistant file in WOF4, so we can't #import
 * it. We need to declare the methods we want instead.
 */
@interface NSObject (PBProjectTypeDeclarations)
- (BOOL)doesMajorVersioning;
- (NSArray *)buildTargets;
@end

@interface NSView (ApplePrivate)
- (void)_recursiveSetDefaultKeyViewLoop;
@end

#ifdef FOUNDATION_STATIC_INLINE
#define YELLOWBOX
#endif
#ifdef NSFoundationVersionNumber10_0
#define COCOA
#endif

static id IvarNamed(id instance, NSString *ivarName)
{
    id ivar;
    object_getInstanceVariable(instance, [ivarName cString], (void **)&ivar);
    return ivar;
}

@implementation RCEnhancePB

+ (void)initialize
{
    [[self alloc] init];
}

- init
{
    NSNotificationCenter *center = [NSNotificationCenter defaultCenter];

#ifdef YELLOWBOX
    // This bugfix is only needed under YB and maybe OSX, I think
    [self addReenableFlushMenuItem];
#endif

#ifndef YELLOWBOX
    // YellowBox versions added this feature
    [self addClearBufferMenuItem];
#endif
    
    // Add "Project Header" attribute to headers
    [center addObserver:self selector:@selector(fileAdded:)
            name:PBFileAddedToProjectNotification object:nil];
    [center addObserver:self selector:@selector(resetBuildPopupState:)
            name:MAKE_BUTTON_NOTIFICATION_NAME object:nil];

    // Add Make popup to build window
    [center addObserver:self selector:@selector(windowOpened:)
            name:NSWindowDidBecomeKeyNotification object:nil];

    return self;
}

/******************************************************************/
// Methods to fix Project Find panel flush problem

- (void)addReenableFlushMenuItem
{
#ifndef YELLOWBOX
// OPENSTEP doesn't have -submenu method
#define submenu target
#endif
    NSMenu *mainMenu  = [[NSApplication sharedApplication] mainMenu];
    NSMenu *toolsMenu = [[mainMenu itemWithTitle:@"Tools"] submenu];
    NSMenu *findMenu  = [[toolsMenu itemWithTitle:@"Project Find"] submenu];
    NSMenuItem *reenableItem;

    reenableItem = [findMenu addItemWithTitle:@"Reenable Flush" action:NULL keyEquivalent:@""];
    [reenableItem setTarget:self];
    [reenableItem setAction:@selector(reenableFlushWindowDamnit:)];
}

- (NSWindow *)projectFindWindow
{
    NSArray *windows = [[NSApplication sharedApplication] windows];
    Class finderClass = NSClassFromString(@"Finder");
    int i, count = [windows count];

    for (i=0; i<count; i++)
    {
        NSWindow *window = [windows objectAtIndex:i];

        if ([window title] &&
            [[window title] rangeOfString:@"Project Find"].length > 0 &&
            [[window delegate] isKindOfClass:finderClass])
        {
            return window;
        }
    }

    return nil;
}

- (void)reenableFlushWindowDamnit:(id)sender
{
    NSWindow *window = [self projectFindWindow];

    while ([window isFlushWindowDisabled])
        [window enableFlushWindow];
}

/******************************************************************/
// Methods to clear out GDB buffer at runtime.  Apple added this in YellowBox,
// but OPENSTEP doesn't have it.

- (void)addClearBufferMenuItem
{
    NSMenu *mainMenu  = [[NSApplication sharedApplication] mainMenu];
    NSMenu *toolsMenu = [[mainMenu itemWithTitle:@"Tools"] submenu];
    NSMenu *launchMenu  = [[toolsMenu itemWithTitle:@"Launcher"] submenu];
    NSMenuItem *clearItem;

    clearItem = [launchMenu addItemWithTitle:@"Clear buffer" action:NULL keyEquivalent:@"k"];
    [clearItem setTarget:self];
    [clearItem setAction:@selector(clearLauncherBuffer:)];
}

- (NSTextView *)launcherTextView
{
    NSApplication *app = [NSApplication sharedApplication];
    Class launcherClass = NSClassFromString(@"Launcher");
    id launcher = nil;

    if ([[[app keyWindow] delegate] isKindOfClass:launcherClass])
        launcher = [[app keyWindow] delegate];
    else if ([[[app mainWindow] delegate] isKindOfClass:launcherClass])
        launcher = [[app mainWindow] delegate];

    return [launcher performSelector:@selector(ttyText)];
}

- (void)clearLauncherBuffer:(id)sender
{
    /*
     * The text is actually an DumbTtyText instance, and responds to
     * -clear. The normal manipulation methods have been overridden, so
     * they can't be used.
     */
    [[self launcherTextView] performSelector:@selector(clear)];
}


- (BOOL)validateMenuItem:(NSMenuItem *)item
{
    if ([item action] == @selector(reenableFlushWindowDamnit:))
    {
        NSWindow *findWindow = [self projectFindWindow];
        return ([findWindow isVisible] && [findWindow isFlushWindowDisabled]);
    }
    if ([item action] == @selector(clearLauncherBuffer:))
    {
        return [self launcherTextView] != nil;
    }

    return YES;
}

/********************************************************************************/
// Auto-set the project header attribute
- (void)fileAdded:(NSNotification *)notification
{
    PBProject *project  = [notification object];
    NSString  *file     = [[notification userInfo] objectForKey:@"file"];
    NSString  *category = [[notification userInfo] objectForKey:@"key"];

    if ([category isEqualToString:PB_HeadersKey])
    {
        PBProject *rootProject = [project nonAggregateRootProject];
        id /*PBProjectType * */ projectType = [rootProject projectType];

        /* We only want to auto-add for frameworks, libraries, etc. */
        if ([projectType doesMajorVersioning])
        {
            NSMutableArray *fileList = [project fileListForKey:PB_PublicHeadersKey];

            if (![fileList containsObject:file])
                [fileList addObject:file];
        }
    }
}

/********************************************************************************/
// Methods to add a popup button with the build targets on the Make window.
// Ugh.  Each project has its own *NSView* that gets placed into the Build
// window, so we have to allocate a popup for each project.  Sigh.

static NSMutableDictionary *popupDictionary = nil;

- (NSPopUpButton *)popupForWindow:(NSWindow *)window create:(BOOL)shouldCreate;
{
    NSValue       *windowKey = [NSValue valueWithPointer:[window contentView]];
    NSPopUpButton *makePopup = [popupDictionary objectForKey:windowKey];

    if (makePopup == nil && shouldCreate)
    {
        id       maker         = [window delegate];
        NSView   *contentView  = [window contentView];
        NSButton *buildButton  = IvarNamed(maker, @"buildButton");
        NSButton *optionButton = IvarNamed(maker, @"optionsButton");
        BOOL     didAutoresize = [contentView autoresizesSubviews];
        NSRect   popupFrame    = NSMakeRect(0, 0, 0.0, 20.0);
        NSRect   contentFrame  = [contentView frame];
        NSRect   buildFrame    = [buildButton convertRect:[buildButton bounds]
                                                            toView:contentView];
        NSRect   optionFrame   = [optionButton convertRect:[optionButton bounds]
                                                             toView:contentView];

        popupFrame.origin.x = buildFrame.origin.x;
        popupFrame.origin.y = NSHeight(contentFrame) - 4.0;
        popupFrame.size.width = NSMaxX(optionFrame) - NSMinX(buildFrame);
        makePopup = [[NSPopUpButton alloc] initWithFrame:popupFrame];
        [makePopup setPullsDown:YES];
        [makePopup setTitle:@"Make:"];
        [makePopup setAutoresizingMask:(NSViewMaxXMargin | NSViewMinYMargin)];
        [makePopup setTarget:self];
        [makePopup setAction:@selector(popupMakeAction:)];
#ifdef COCOA
        [makePopup setFont:[NSFont labelFontOfSize:11.0]];
        [[makePopup cell] setControlSize:NSSmallControlSize];
#endif

        [window disableFlushWindow];
        [contentView setAutoresizesSubviews:NO];
        contentFrame.size.height += popupFrame.size.height;
        [contentView setFrame:contentFrame];
        [contentView addSubview:makePopup];
        [contentView setAutoresizesSubviews:didAutoresize];
        contentFrame.size.height -= popupFrame.size.height;
        contentFrame.size.width += 3.0;
        [contentView setFrame:contentFrame];
        [contentView _recursiveSetDefaultKeyViewLoop];		// Is there public api for this?
        [contentView setNeedsDisplay:YES];
        [makePopup setNextKeyView:buildButton];
        [window enableFlushWindow];

        if (popupDictionary == nil) popupDictionary = [[NSMutableDictionary alloc] init];
        [popupDictionary setObject:makePopup forKey:windowKey];
        [makePopup release];
    }

    return makePopup;
}

- (void)resetBuildPopupState:(NSNotification *)note
{
    NSButton *button = [note object];
    id topView = button;

    // Find the popup via the button's contentView. [button window] may be nil, so do it the hard way.
    while ([[topView superview] superview])
        topView = [topView superview];
    [[popupDictionary objectForKey:[NSValue valueWithPointer:topView]] setEnabled:[button isEnabled]];
}

- (NSArray *)enhanceBuildTargets
{
    NSArray *targets = [[NSUserDefaults standardUserDefaults] arrayForKey:@"RCEnhanceBuildTargets"];
    if (targets == nil)
        targets = [@"(install_debug, install_unstripped, install_profile)" propertyList];
    return targets;
}

static void _AddPopupItems(NSPopUpButton *button, NSArray *targets)
{
    int i, count = [targets count];
    for (i=0; i<count; i++) {
        NSString *item = [targets objectAtIndex:i];
        if ([button indexOfItemWithTitle:item] < 0)
            [button addItemWithTitle:item];
    }
}

- (void)resetPopupInMakeWindow:(NSWindow *)window
{
    id <PBBundleHost> app = [NSApplication sharedApplication];
    NSString          *projectPath = [app selectedProject];
    NSString          *canonFile = [projectPath stringByAppendingPathComponent:@"PB.project"];
    PBProject         *project = [PBProject projectForCanonicalFile:canonFile];
    NSArray           *extraTargets = [project objectForKey:@"TARGETS"];
    NSPopUpButton     *makePopup;

    makePopup = [self popupForWindow:window create:YES];

    [makePopup removeAllItems];
    [makePopup addItemWithTitle:@"Make:"];
    _AddPopupItems(makePopup, [(id)[project projectType] buildTargets]);
    _AddPopupItems(makePopup, extraTargets);
    _AddPopupItems(makePopup, [self enhanceBuildTargets]);
    [makePopup addItemWithTitle:@"<Default>"];
}

- (void)windowOpened:(NSNotification *)notification
{
    NSWindow *window = [notification object];
    Class    makerClass = NSClassFromString(@"Maker");

    if ([window title] &&
        [[window title] rangeOfString:@"Project Build"].length > 0 &&
        [[window delegate] isKindOfClass:makerClass])
    {
        [self resetPopupInMakeWindow:window];
    }
}
    
- (void)popupMakeAction:(id)sender
{
    if ([sender indexOfSelectedItem] > 0)
    {
        id            maker      = [[sender window] delegate];
        NSPopUpButton *realPopup = IvarNamed(maker, @"targetPopUp");
        NSString      *title     = [sender titleOfSelectedItem];

        [realPopup addItemWithTitle:title];  //make sure it exists
        [realPopup selectItemWithTitle:title];
        [maker performSelector:@selector(switchTarget:) withObject:realPopup];
        [maker performSelector:@selector(build)];
        [sender setEnabled:NO];
    }
}

@end

/*
 * Evil hack, stolen from PBButtons -- whenever the enabled status of PB's
 * build buttons change, update our popup the same way. A cleaner way may
 * be to listen for all the build stop notifications, and check the PB
 * button's status, and update accordingly. We'd have to see if only one
 * button is enabled, since during a build one or the other is disabled,
 * but not both. We'd also have to have a second dictionary of
 * project->popup, since I don't know how to find the content view given
 * just the project. This hack works, so leave it.
 */
@implementation NSButton (BuildEnablingHack)

- (void)setEnabled:(BOOL)flag
{
    [super setEnabled:flag];

    if ([self action] == @selector(make:) || [self action] == @selector(makeClean:))
    {
        NSNotificationCenter *center = [NSNotificationCenter defaultCenter];
        [center postNotificationName:MAKE_BUTTON_NOTIFICATION_NAME object:self];
    }

    /* We don't need this ourselves, but if PBButtons is installed it does */
    if ([[[self image] name] isEqualToString:@"processes"]
        && [[[self window] delegate] isKindOfClass:NSClassFromString(@"Launcher")])
    {
        NSNotificationCenter *center = [NSNotificationCenter defaultCenter];
        [center postNotificationName:LAUNCH_BUTTON_NOTIFICATION_NAME object:self];
    }
}

@end
